.report-hint {
            margin-bottom: 1rem;
            font-size: 0.95rem;
            color: #2d3748;
        }
        .modal {
            position: fixed;
            inset: 0;
            display: none;
            align-items: center;
            justify-content: center;
            padding: 1.5rem;
            z-index: 1000;
        }
        .modal.is-open {
            display: flex;
        }
        .modal__backdrop {
            position: absolute;
            inset: 0;
            background: rgba(0, 0, 0, 0.45);
        }
        .modal__content {
            position: relative;
            background: #ffffff;
            border-radius: 0.75rem;
            width: min(1100px, 94vw);
            height: 90vh;
            max-height: 90vh;
            display: flex;
            flex-direction: column;
            box-shadow: 0 12px 30px rgba(0, 0, 0, 0.2);
            overflow: hidden;
            z-index: 1;
        }
        .modal__body {
            display: flex;
            flex-direction: column;
            background: #ffffff;
            flex: 1 1 auto;
            min-height: 0;
            overflow: hidden;
        }
        .modal__header,
        .modal__footer {
            display: flex;
            justify-content: space-between;
            align-items: center;
            gap: 1rem;
            padding: 0.9rem 1.2rem;
            background: #f7fafc;
            border-bottom: 1px solid #e2e8f0;
        }
        .modal__footer {
            border-top: 1px solid #e2e8f0;
            border-bottom: none;
        }
        .modal__title {
            font-size: 1rem;
            font-weight: 600;
            color: #1a202c;
        }
        .modal__close {
            border: none;
            background: #e2e8f0;
            color: #1a202c;
            padding: 0.4rem 0.8rem;
            border-radius: 0.4rem;
            cursor: pointer;
        }
        .modal__frame {
            width: 100%;
            height: 100%;
            border: none;
            background: #ffffff;
            flex: 1 1 60%;
            min-height: 180px;
        }
        .modal__divider {
            height: 16px;
            min-height: 16px;
            flex: 0 0 16px;
            background: #e2e8f0;
            cursor: row-resize;
            border-top: 1px solid #cbd5e0;
            border-bottom: 1px solid #cbd5e0;
            touch-action: none;
            user-select: none;
        }
        .modal__igv {
            display: flex;
            flex-direction: column;
            gap: 0.5rem;
            padding: 0.75rem 1rem 1rem;
            border-top: 1px solid #e2e8f0;
            background: #ffffff;
            flex: 1 1 40%;
            min-height: 180px;
        }
        .modal__igv-title {
            font-size: 0.9rem;
            font-weight: 600;
            color: #1a202c;
        }
        .modal__igv-status {
            font-size: 0.85rem;
            color: #4a5568;
        }
        .modal__igv-view {
            width: 100%;
            height: 100%;
            border: 1px solid #e2e8f0;
            border-radius: 0.5rem;
            overflow: visible;
            background: #f7fafc;
        }
        .igv-popover {
            z-index: 2000;
            pointer-events: auto;
        }
        .modal__link {
            text-decoration: none;
            color: #3182ce;
            font-weight: 600;
        }
