"""
Error detection and correction utilities.

This module handles:
- Machine learning-based sequence error detection
- Homopolymer region error identification
- Strand bias detection and correction
"""
